/*
    clock.c (Nethack-Player/Recorder) Version 1.0
    Copyright (C) 2000 by Helge Dennhardt
    <Helge.Dennhardt@gmx.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   With "clock.h" and "clock.c" we have a (simulated) clock, measuring time
   in multiples of 1/100th of a second, beginning from the start of the game.

   This clock is simulated, which means that it gets no timer from the 
   system, and there is no timer signalling or interruption and such stuff.

   Instead, whenever asked, it gets the current system time and then calculates
   the readings of the (simulated) clock. This calculation is based on a 
   starting time index (which is stored) given at a certain system time
   (which is stored) and a given clock speed (which is also stored). 
*/


#include <stdlib.h>   /* EXIT_FAILURE */
#include <sys/time.h> /* gettimeofday, timeval */
#include <errno.h>    /* error, errno */
#include "clock.h"    /* TIME, SPEED */

/* These three variables hold the status of the (simulated) clock. See StartClock() for details.*/
static TIME Time;
static int  Speed;
static struct timeval SystemTime;


/* Sets the (simulated) clock to <Time>. This clock will run <Speed>
   times faster than real time. */ 
void StartClock(TIME Time_, SPEED Speed_)
{
  /* Store the initial time of the simulated clock */
  Time = Time_;

  /* Store the Speed of the simulated clock */
  Speed = Speed_;

  /* Get current system time and store it*/
  if (gettimeofday(&SystemTime,0)<0) error(EXIT_FAILURE, errno, 0);
}


/* Reads out the (simulated) clock. All time is measured in 1/100th of a second.*/
TIME GetTime()
{
  /* First read out the current system time. */
  struct timeval CurrentTime;
  if (gettimeofday(&CurrentTime,0)<0) error(EXIT_FAILURE, errno, 0);
  
  /* Then calculate the readings of the (simulated) clock. 
     This is done by first calculating the elapsed seconds multiplied by 100 to get 1/100th 
     of seconds, then adding elapsed microseconds multiplied by 10000 to get 1/100th of seconds. 
     This elapsed system time is then speeded up by multiplying it with the Speed parameter 
     and then added to the time base of our clock. */
  return(((CurrentTime.tv_sec-SystemTime.tv_sec)*100
         +(CurrentTime.tv_usec-SystemTime.tv_usec)/10000
        )*Speed+Time);
}


/* Reads out the current speed multiplicator of the (simulated) clock.*/
SPEED GetSpeed()
{
  return Speed;
}













