/*
    Nethack-Player Version 1.0
    Copyright (C) 2000 by Helge Dennhardt
    <Helge.Dennhardt@gmx.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   How it works:

   The Nethack-Player reads character by character from the movie file and
   sends it to the standard output. This is done at high speed.

   With "clock.h" and "clock.c" we have a (simulated) clock, measuring time
   in multiples of 1/100th of a second, beginning from the start of the game.
   Whenever the recorder reads a null byte with a following time stamp from
   the movie file, it waits until this time index is reached, before continuing.

   This (simulated) clock could run at different speed, which can be altered 
   with the option -s or by typing 0-9 while running.

   This (simluated) clock could start at a time index you want, which can
   be set with the option -t. This means that the movie is shown from the
   beginning at high speed until the given time index is reached and then
   continued normally with the demanded speed.
 
   This showing the movie always from the start is necessary to get the screen
   content correct. On my machine (Pentium II) this takes virtually no time,
   but on slower machines this could be annoying. I don't know a possibility
   to avoid this without storing the whole screen content, which would make
   the movie files even larger than thy are now. 

   When the movie ends or is interrupted, it will give a short message
   at which time index we are now. This makes it easier to resume.

   Enjoy!
*/

#include <stdlib.h>  /* printf, sprintf, sscanf, exit, EXIT_SUCCES, EXIT_FAILURE */
#include <fcntl.h>   /* open, close, read, write, ssize_t, O_RONLY */
#include <errno.h>   /* error, errno */
#include "clock.h"   /* Startclock(), ReadClock(), TIME, SPEED */
#include <termios.h> /* termios, ICANON, VMIN, VTIME, TCSANOW, tcgetattr, tcsetattr, isatty */
#include <unistd.h>  /* STDIN_FILENO, STDOUT_FILENO */

/* Insert here the escape code of your terminal to clear the screen. You can get this by
   typing "clear >Text" and then reading the file "Text" f.e. with emacs */
#define CLEARSCREEN printf("[H[2J"); 

static SPEED Speed;         /* The current speed of the (simulated) clock. */
static TIME  Time;          /* The time we have to wait for. This could be the starting time given
                               as a parameter or the last read time stamp from the movie file. */
static char* MovieFileName; /* The name of the movie file to play */


struct termios SavedAttrib; /* All attributes of the current terminal are stored here to be restored
                               at the end of the program. */

/* At the beginning of NH_Player() the attributes of the current terminal are changed. This routine
   restores the previous state before exiting the program. Normally the shell should do this, but
   some don't. */
void ResetInputMode()
{
  if (tcsetattr(STDIN_FILENO, TCSANOW, &SavedAttrib)<0)
    error("EXIT_FAILURE, 0, Error - Could not reset input mode of STDIN!");
}


/* When the program ends or is interrupted, give out a message with the current time, so that it is
   easier to estimate times for the -t option. */ 
void PrintTime()
{
  char SpeedBuffer[100];
  char TimeBuffer[100];

  /* If the speed of the (simulated) clock is not normal, incorporate this in the message. */
  Speed = GetSpeed();
  if (Speed==1)
    sprintf(SpeedBuffer, "");
  else
    sprintf(SpeedBuffer, "-s%c ", '0'+Speed);

  /* Compose the time string in the form "days.hours.minutes.seconds". */ 
  Time = GetTime();
  sprintf(TimeBuffer, "%d.%d.%d.%d", (Time/8640000), (Time/360000)%24, (Time/6000)%60, (Time/100)%60);  

  /* Finally, print out the whole message. Use different messages for ending/interrupting the movie. */
  if (MovieFileName==0)
    printf("\n\nMovie ended at time %s\n\n", TimeBuffer);
  else
    printf("\n\nMovie interrupted. Continue with \"nh_player %s-t%s %s\".\n\n", SpeedBuffer, TimeBuffer,
      MovieFileName);
}

void NH_Player() /* Main part. */
{
  int     MovieFile;          /* File descriptor of the movie file */
  char    Byte;               /* Last read byte from the movie file */
  ssize_t BytesRead;          /* Number of Bytes read from the movie file */
  struct  termios TermAttrib; /* New attributes of the current terminal */

  /* Be sure that the standard input (used for the keyboard commands) is really a terminal. */
  if (!isatty(STDIN_FILENO)) 
    error(EXIT_FAILURE, 0, "ERROR - STDIN is not a terminal!");

  /* Save all attributes of the current terminal */
  if (tcgetattr(STDIN_FILENO, &SavedAttrib)<0) 
    error(EXIT_FAILURE, 0, "Error - Could not read input mode of STDIN!");

  /* and restore them when the program ends or is interrupted. */
  atexit(ResetInputMode);

  /* Now change the attributes of the current terminal as is needed for the keyboard commands:
       ~ICANON changes line oriented input to character oriented input.
       ~ECHO turns off the keyboard echo.
       VMIN=0, VTIME=0 tells the read() function not to wait for keyboard input. */
  if (tcgetattr(STDIN_FILENO, &TermAttrib)<0) 
    error(EXIT_FAILURE, 0, "Error - Could not read input mode of STDIN!");
  TermAttrib.c_lflag &= ~(ICANON|ECHO);
  TermAttrib.c_cc[VMIN]=0;
  TermAttrib.c_cc[VTIME]=0;
  if (tcsetattr(STDIN_FILENO, TCSANOW, &TermAttrib)<0)
    error(EXIT_FAILURE, 0, "Error - Could not change input mode of STDIN!");

  /* Now open the movie file in read-only mode. */
  MovieFile=open(MovieFileName, O_RDONLY);
  if (MovieFile<0) error(EXIT_FAILURE, errno, MovieFileName);

  /* Start time measuring with the (simulated) clock. 
     Set time to zero or to the time specified with the -t option.
     Set speed to normal or the speed specified with the -s option. */
  StartClock(Time, Speed);

  /* Print out current Time when program is ended or interrupted. */
  atexit(PrintTime);

  /* Main loop. Read byte by byte from the movie file. */
  while (1)
  {  
    /* Read a byte from standard input. */
    BytesRead = read(MovieFile, &Byte, 1);
    if (BytesRead<0) error(EXIT_FAILURE, errno, MovieFileName); 

    /* Exit at end of movie file. */
    if (BytesRead==0) break;          

    /* if this byte is zero, a time stamp follows and we have to wait until this time. */
    if (Byte==0)
    {
      BytesRead = read(MovieFile, &Time, sizeof(Time));
      if (BytesRead<sizeof(Time)) error(EXIT_FAILURE, errno, MovieFileName);

      /* While waiting, watch the keyboard for keyboard commands. */ 
      while (GetTime()<Time)
      {
        if (read(STDIN_FILENO, &Byte, 1)==1) 
	{
          /* If an interruption is requested, close the movie file and exit. */
	  if ((Byte=='q')||(Byte=='Q'))
          {
            if (close(MovieFile)<0) error(EXIT_FAILURE, errno, MovieFileName);
            CLEARSCREEN;
            exit(EXIT_SUCCESS);
          }
          /* If a speed change is requested, change the settings of the (simulated) clock. */
          if ((Byte>='0')&&(Byte<='9'))
	  {
            Speed = Byte-'0';
            StartClock(GetTime(), Speed);
          }
        }
      }/* End of waiting loop. */
    }
    /* If there is no time stamp, but a normal character, send this byte to standard output. */
    else    
      if (write(STDOUT_FILENO, &Byte, 1)<0) error(EXIT_FAILURE, errno, "stdout");

  }/* End of main loop. */

  /* Close movie file. */
  if (close(MovieFile)<0) error(EXIT_FAILURE, errno, MovieFileName);

  /* Clear the screen to get rid of all  the rubbish. */
  CLEARSCREEN;

  /* This is to signal PrintTime() that the end of the movie is reached and the message has
     to bet altered a bit. */ 
  MovieFileName = 0;
}


void Usage() /* Print out the help message */
{
  printf("\nThis is the Nethack-Player Version 1.0\n");
  printf("   (c) 2000 by Helge Dennhardt\n\n");
  printf("SYNTAX:\n");
  printf("  nh_player [ -s<speed> ] [ -t<days>.<hours>.<minutes>.<seconds> ]\n");
  printf("            <movie-filename>\n\n");
  printf("DESCRIPTION:\n");
  printf("  With it  you can  play the  movie-files  generated by  the Nethack-Recorder\n");
  printf("  Version  1.0  (type \"nh_recorder\"  to get a  short help  message).  Enjoy\n");
  printf("  watching your old Nethack-games!\n\n");
  printf("PARAMETERS:\n");
  printf(" -s<speed>        Here you  adjust the speed in which the movies  are played.\n");
  printf("                  Only one digit  is allowed, so the speed can range from -s0\n");
  printf("                  (total stop) over -s1  (normal speed) up to -s9 (nine times\n"); 
  printf("                  normal speed). Default is -s1.\n"); 
  printf(" -t<days>.<hours>.<minutes>.<seconds>\n");
  printf("                  Though several  speeds are supported, for technical reasons\n");
  printf("                  there  is no rewind. Nevertheless, you can stop  the player\n");
  printf("                  and with  this option restart it at a time index you  want.\n");
  printf("                  For  example -t0.1.20.0 will start the  movie one  hour and\n");
  printf("                  twenty minutes after its start. Default is -t0.0.0.0\n");
  printf(" <movie-filename> The filename of the movie you want to play.\n\n");
  printf("KEYS:\n");
  printf(" 0-9              You can change the speed of the movie at any time by\n");
  printf("                  hitting the appropriate number keys.\n");
  printf(" q,Q              Quits the program.\n\n");
}


int main(int argc, char** argv) /* Analysing parameters */
{
  int   nParameter;    /* Actual parameter number to be analysed, */
  char* pParameter;    /* and a pointer to it. */

  int   Days;          /* Variables which hold the information given with the -t option. */
  int   Hours;
  int   Minutes;
  int   Seconds;

  /* When invoked with no parameters, give a small help message and exit. */
  if (argc==1)
  {
    Usage();
    exit(EXIT_FAILURE);
  }
   
  /* Otherwise set all parameters to default values. */
  Speed = 1;
  Time  = 0;
  MovieFileName = 0;

  /* Then analyse parameters one by one. */
  for (nParameter=1; nParameter<argc; nParameter++)
  {
    pParameter=argv[nParameter];

    /* If a parameter starts with "-" then it is an option */
    if (pParameter[0]=='-')
      switch(pParameter[1])
      {
      case 's': /* For parameter "s" first check the syntax */
                if ((pParameter[2]<'0') || (pParameter[2]>'9') || (pParameter[3]!=0))
                  error(EXIT_FAILURE, 0, "Error - Allowed speed ranges from -s0 to -s9 !");

	        /* Then set the speed variable to the demanded value. */
                Speed = pParameter[2]-'0';
                break;

      case 't': /* For parameter "t" try to read the time index. */ 
                if (sscanf(pParameter, "-t%d.%d.%d.%d\0", &Days, &Hours, &Minutes, &Seconds)<4)
                  error(EXIT_FAILURE, 0, "Error - Time format is -t<days>.<hours>.<minutes>.<seconds> !");

                /* Then set the time variable to the demanded value. Time is measured in 1/100th of
                   a second. */
                Time = ((((Days*24)+Hours)*60+Minutes)*60+Seconds)*100;
                break;

      default:  /* For other parameters than "s" or "t" give an error message and exit. */
                printf("nh_player: Error - Unknown option '%c' !\n", pParameter[1]);
                exit(EXIT_FAILURE);
      }  

    /* If the parameter doesen't start with "-", then it is the movie filename. */
    else
    {
      /* First check if the filename is already set */
      if (MovieFileName!=0)
      {
        printf("nh_player: Error - Two movie files \"%s\" and \"%s\" !\n", MovieFileName, pParameter);
        exit(EXIT_FAILURE);
      }
        
      /* If not, set it now. */
      MovieFileName = pParameter;
    }
  }
 
  /* The movie filename is obligatory */
  if (MovieFileName==0)
    error(EXIT_FAILURE, 0, "Error - Name of movie file is missing !");

  /* All parameters are set. Now start business. */
  NH_Player(); 

  /* Finally, end the program successfully. */
  return EXIT_SUCCESS;
}















