/*
    Nethack-Recorder Version 1.0
    Copyright (C) 2000 by Helge Dennhardt
    <Helge.Dennhardt@gmx.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   How it works:

   The Nethack-Recorder transports every character from standard input to 
   the standard output, so that it doesn't interfere with the actual Nethack
   game. In addition, it stores every character in the movie file. The name 
   of this movie file is given as a parameter.
   
   With "clock.h" and "clock.c" we have a (simulated) clock, measuring time
   in multiples of 1/100th of a second, beginning from the start of the game.
   Whenever the recorder has to wait longer than 1/10th of a second for a
   character, it stores a null byte, the actual time and then the charcter
   it just recieved.  

   When the program terminates, it makes sure that the movie file ends with 
   such a timestamp. When you resume your saved Nethack game,
   the recorder will look up the actual time at the end of the movie
   file and then append to it.

   Enjoy!
*/


#include <stdlib.h> /* printf, exit, EXIT_FAILURE */
#include <fcntl.h>  /* open, close, read, write, ssize_t, O_RDWR, O_WRONLY, O_CREAT, O_APPEND */
#include <errno.h>  /* error, errno */
#include <unistd.h> /* STDIN_FILNO, STDOUT_FILENO, SEEK_END */
#include "clock.h"  /* StartClock(), GetTime(), TIME, SPEED */


/* This is the name of the movie file to be recorded. It is the only parameter. */
static char* MovieFileName;


void NH_Recorder() /* Main part */
{
  int        MovieFile;    /* File descriptor of the movie file */
  char       Byte;         /* Last read byte from standard input */
  ssize_t    BytesRead;    /* Number of Bytes read from standard input */
  TIME       TimeStamp;    /* For saving the current time of the (simulated) clock */
  const char Zero=0;       /* This is always 0. */

  /* First, try to append to the movie file. */ 
  MovieFile = open(MovieFileName, O_RDWR|O_APPEND, 0666);

  /* If it is non-existent, create it with access mode 0666 and start at zero time.*/
  if (MovieFile<0)
  {
    MovieFile = open(MovieFileName, O_WRONLY|O_CREAT, 0666);
    if (MovieFile<0) error(EXIT_FAILURE, errno, MovieFileName);
    TimeStamp = 0;
  }
  /* Otherwise append to it. To continue time measuring, read the last time stamp from
     the end of the file */
  else
  {
    lseek(MovieFile, -sizeof(TimeStamp), SEEK_END);
    BytesRead= read(MovieFile, &TimeStamp, sizeof(TimeStamp));
    if (BytesRead<sizeof(TimeStamp)) error(EXIT_FAILURE, errno, MovieFileName);
  }

  /* Start time measuring. Set the (simulated) clock to TimeStamp at normal speed. */
  StartClock(TimeStamp, 1);

  /* Main loop. Read byte by byte from standard input. */
  while (1)
  {  
    /* Before reading the first byte and thus giving up control to Nethack, save the current time. */
    TimeStamp = GetTime();
  
    /* Read a byte from standard input. */
    BytesRead = read(STDIN_FILENO, &Byte, 1);
    if (BytesRead<0) error(EXIT_FAILURE, errno, MovieFileName); 

    /* If this took longer than 1/10th of a second, Nethack has waited for user input.
       Notify this and write the current time to the movie file. 
       Do this also when transmission ends, so every movie file ends with a time stamp.*/ 
    if ((GetTime()-TimeStamp>10)||(BytesRead==0))
    {
      /* First write a single zero-byte to notify that a timestamp is following. */
      if (write(MovieFile, &Zero, 1)<0) error(EXIT_FAILURE, errno, MovieFileName);

      /* Then write the timestamp */
      if (write(MovieFile, &TimeStamp, sizeof(TimeStamp))<0) error(EXIT_FAILURE, errno, MovieFileName); 
    }

    /* Exit at end of transmission. */
    if (BytesRead==0) break;          

    /* Write to standard output. */
    if (write(STDOUT_FILENO, &Byte, 1)<0) error(EXIT_FAILURE, errno, "stdout");

    /* Write to movie file. */
    if (write(MovieFile, &Byte, 1)<0) error(EXIT_FAILURE, errno, MovieFileName);    

  } /* End of main loop */
    
  /* Close movie file. */
  if (close(MovieFile)<0) error(EXIT_FAILURE, errno, MovieFileName);
}


void Usage() /* Print out the help message */
{
  printf("\nThis is the Nethack-Recorder Version 1.0\n");
  printf("   (C) 2000 by Helge Dennhardt\n\n");
  printf("SYNTAX:\n");
  printf("  nethack | nh_recorder <movie-filename>\n\n");
  printf("DESCRIPTION:\n");
  printf("  With it you can record your Nethack-game into a file, which can be replayed\n");
  printf("  later with Nethack-Player Version 1.0 (type \"nh_player\" to get a short help\n");
  printf("  message). Simply redirect the standard  output into the Nethack-recorder as\n");
  printf("  shown in the above syntax. Enjoy watching your old Nethack-games!\n\n");
  printf("PARAMETER:\n");
  printf("  <movie-filename> is the name of the file, in which the movie-information is\n");
  printf("  to be stored. For this I suggest \".nh\" as file  extension for Nethack movie\n");
  printf("  files. In  case a file  with this  name already  exists,  the recorder will\n");
  printf("  append to it.\n\n");
  printf("NOTE:\n");
  printf("  I beg you:  PLEASE, DON'T POST YOUR MOVIES!  They are too big.  People will\n");
  printf("  get annoyed.  If you want to share your  Nethack adventures, mail them, put\n");
  printf("  them on a website for  downloading or  post them in the alt.binaries.* tree\n");
  printf("  with only a small note in a Nethack newsgroup. Thank you.\n\n");
}


int main(int argc, char** argv) /* Analysing all parameters */
{
  /* When invoked with no or too many parameters, give a small help message and exit. */
  if (argc!=2) 
  { 
    Usage();
    exit(EXIT_FAILURE);
  } 

  /* Otherwise get the movie file name */
  MovieFileName=argv[1];

  /* and start business. */
  NH_Recorder(); 

  /* Finally end the program. */
  return EXIT_SUCCESS;
}














